"""Class to initialise loggers to screen and files"""

import logging
import os

from logging.handlers import RotatingFileHandler

def initialize_logger(output_dir, screenlevel, debug_log=None):
    """Class to initialise loggers to screen and files"""
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)

    # create console handler and set level to info
    handler = logging.StreamHandler()
    handler.setLevel(screenlevel)
    formatter = logging.Formatter("%(name)s - %(levelname)s - %(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    # create error file handler and set level to error
    handler = logging.FileHandler(os.path.join(output_dir, "error.log"), "w", encoding=None, delay="true")
    handler.setLevel(logging.WARN)
    formatter = logging.Formatter("%(asctime)-15s %(levelname)s - %(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    #a is append, w is write
    if debug_log != None:
        #create debug file handler and set level to debug
        handler = RotatingFileHandler(os.path.join(output_dir, "all.log"), mode='a', maxBytes=5*1024*1024,
                                     backupCount=2, encoding=None, delay=0)
        #handler = logging.FileHandler(os.path.join(output_dir, "all.log"),"w", encoding=None, delay="true")
        handler.setLevel(logging.DEBUG)
        formatter = logging.Formatter("%(asctime)-15s %(levelname)s - %(message)s")
        handler.setFormatter(formatter)
        logger.addHandler(handler)



"""Handles Heatmiser network of devices and adaptor

also loads adaptor and keeps track of Heatmiser devices

Ian Horsley 2018
"""

import os
import logging
import sys

# Import our own stuff
from devices import HeatmiserDevice, HeatmiserBroadcastDevice
from adaptor import HeatmiserAdaptor
import setup as hms

class HeatmiserNetwork(object):
    """Class that connects a set of devices (from configuration) and an adpator."""
    ### stat list setup

    def __init__(self, configfile = None):
        
        # Select default configuration file if none provided
        if configfile is None:
            self._module_path = os.path.abspath(os.path.dirname(__file__))
            configfile = os.path.join(self._module_path, "hmcontroller.conf")
          
        # Initialize controller setup
        try:
            self._setup = hms.HeatmiserControllerFileSetup(configfile)
            settings = self._setup.settings
        except hms.HeatmiserControllerSetupInitError as err:
            logging.error(err)
            sys.exit("Unable to load configuration file: " + configfile)
        
        # Initialize and connect to heatmiser network, probably through serial port
        self.adaptor = HeatmiserAdaptor(self._setup)
        self.adaptor.connect()
        
        self._set_stat_list(settings['devices'], settings['devicesgeneral'])
        
        # Create a broadcast device
        setattr(self, "All", HeatmiserBroadcastDevice(self.adaptor, "Broadcast to All", self.controllers))
        self._current = self.All
      
    def _set_stat_list(self, statlist, generalsettings):
        """Store list of devives and create objects for each"""
        self._statlist = statlist
        self._statnum = len(self._statlist)

        self.controllers = range(self._statnum)
        for name, controllersettings in statlist.iteritems():
            if hasattr(self, name):
                print "error duplicate stat name"
            else:
                setattr(self, name, HeatmiserDevice(self.adaptor, controllersettings, generalsettings))
                setattr(getattr(self, name), 'name', name) #make name avaliable when accessing by id
                self.controllers[controllersettings['display_order']-1] = getattr(self, name)

        self._current = self.controllers[0]
      
    def get_stat_address(self, shortname):
        """Get network address from device name."""
        if isinstance(shortname, basestring):
            return self._statlist[shortname]['address']
        else:
            return shortname

    def set_current_controller_by_name(self, name):
        """Set the current device by name"""
        self._current = getattr(self, name)
        
    def set_current_controller_by_index(self, index):
        """Set the current device by id"""
        self._current = self.controllers[index]
        
    def get_controller_by_name(self, name):
        """Get device if from name"""
        return getattr(self, name)
        
    def run_method_on_all(self, method, *args, **kwargs):
        """Run a method on all devices"""
        results = []
        for obj in self.controllers:
            results.append(getattr(obj, method)(*args, **kwargs))
        return results



"""User interface to setup the contoller."""

import logging
import os
from configobj import ConfigObj
from validate import Validator

from .exceptions import HeatmiserControllerSetupInitError

# The settings attribute stores the settings of the hub. It is a
# dictionary with the following keys:

        # 'controller': a dictionary containing general settings
        # 'serial': a dictionary containing the serial port settings
        # 'devices': a dictionary containing the configuration of each remote device
        # 'setup':

        # The controller settings are:
        # 'loglevel': the logging level

        # ###interfacers and reporters are dictionaries with the following keys:
        # ###'Type': class name
        # ###'init_settings': dictionary with initialization settings
        # ###'runtimesettings': dictionary with runtime settings
        # ###Initialization and runtime settings depend on the interfacer and
        # ###reporter type.

# The run() method is supposed to be run regularly by the instantiator, to
# perform regular communication tasks.

# The check_settings() method is run regularly as well. It checks the settings
# and returns True is settings were changed.

# This almost empty class is meant to be inherited by subclasses specific to
# each setup."""

class HeatmiserControllerSetup(object):
    """Inherited base class"""
    def __init__(self):
        # Initialize logger
        self._log = logging.getLogger("HeatmiserController")

        # Initialize settings
        self.settings = None

    def run(self):
        """Run in background.

        To be implemented in child class.

        """
        pass

    def check_settings(self):
        """Check settings
        
        Update attribute settings and return True if modified.
        
        To be implemented in child class.
        
        """

class HeatmiserControllerFileSetup(HeatmiserControllerSetup):
    """Handles importing confiugration from file"""
    def __init__(self, filename):
        
        # Initialization
        super(HeatmiserControllerFileSetup, self).__init__()

        self._module_path = os.path.abspath(os.path.dirname(__file__))
        specpath = os.path.join(self._module_path, "hmcontroller.spec")
        
        # List of expected sections
        self._sections = ['controller', 'serial', 'devices', 'setup']
            
        # Initialize attribute settings as a ConfigObj instance
        logging.debug("Loading %s and checking against %s"%(filename, specpath))
        try:
            self.settings = ConfigObj(filename, file_error=True, configspec=specpath)
            self._validator = Validator()
        except IOError as err:
            raise HeatmiserControllerSetupInitError(err)
        except SyntaxError as err:
            raise HeatmiserControllerSetupInitError(
                'Error parsing config file \"%s\": ' % filename + str(err))
        except KeyError as err:
            raise HeatmiserControllerSetupInitError(
                'Configuration file error - section missing: ' + str(err))
        
        #check settings and add any default values
        self._check_settings()
        
        # Initialize update timestamps
        self._settings_update_timestamp = 0
        
        # Load use configured variables
        for name, value in self.settings['setup'].iteritems():
            setattr(self, '_c_'+name, value)

        # create a timeout message if time out is set (>0)
        self.retry_msg = " Retry in " + str(self._c_retry_time_interval) + " seconds" if self._c_retry_time_interval <= 0 else ""

    # def reload_settings(self):
        # """Reload and check settings

        # Update attribute settings and return True if modified. Return False if failed to load new settings. Return None if nothing new or not checked

        # """
        
        # # Check settings only once per second
        # now = time.time()
        # if now - self._settings_update_timestamp < 0:
            # return
        # # Update timestamp
        # self._settings_update_timestamp = now + self._c_check_time_interval
        
        # # Backup settings
        # settings = dict(self.settings)
        
        # # Get settings from file
        # try:
            # self.settings.reload()
        # except IOError as err:
            # self._log.warning('Could not get settings: ' + str(err) + self.retry_msg)
            # self._settings_update_timestamp = now + self._c_retry_time_interval
            # return
        # except SyntaxError as err:
            # self._log.warning('Could not get settings: ' +
                              # 'Error parsing config file: ' + str(err) + self.retry_msg)
            # self._settings_update_timestamp = now + self._c_retry_time_interval
            # return
        # except Exception:
            # import traceback
            # self._log.warning("Couldn't get settings, Exception: " +
                              # traceback.format_exc() + self.retry_msg)
            # self._settings_update_timestamp = now + self._c_retry_time_interval
            # return

        # if self.settings != settings:
            # try:
                # self._check_settings()
            # except (ValueError, KeyError):
                # self.settings = settings
                # return False
            # return True
            
    def _check_settings(self):
        """Function validates configuration against specification."""
        try:
            # Validate the configuration file and copy any missing defaults
            returnval = self.settings.validate(self._validator, preserve_errors=True, copy=True)
            if not returnval is True:
                raise ValueError("failed validation of config file %s"%str(returnval))

            # Check the settings file sections exist
            for name in self._sections:
                if not name in self.settings:
                    raise KeyError("Section %s not defined"%name)
        except (ValueError, KeyError) as err:
            logging.warning("Configuration parse failed : " + str(err))
            raise


            






"""Heatmiser Device Classes

Modules handle all the DCB and fields for each device on the Heatmiser network

Ian Horsley 2018
"""

#read = return local if not to old, otherwise gets
#get = goes to network to get
#each field should it's own maximum age

import logging
import time
import serial
import copy

from hm_constants import *
from .exceptions import HeatmiserResponseError, HeatmiserControllerTimeError
from schedule_functions import SchedulerDayHeat, SchedulerWeekHeat, SchedulerDayWater, SchedulerWeekWater, SCH_ENT_TEMP
from decorators import ListWrapperClass, run_function_on_all
from operator import itemgetter

class HeatmiserDevice(object):
    """General device class for thermostats"""
    ## Variables used by code
    lastreadtime = 0 #records last time of a successful read

    ## Initialisation functions and low level functions
    def __init__(self, adaptor, devicesettings, generalsettings=None):
        #address, protocol, short_name, long_name, model, mode
        self._adaptor = adaptor

        self.water_schedule = None

        #initialise data structures
        self._uniquetodcb = []
        self._fieldsvalid = []
        self._buildfieldtables()
        self.data = dict.fromkeys(self._fieldnametonum.keys(), None)
        self.floorlimiting = None
        self.datareadtime = dict.fromkeys(self._fieldnametonum.keys(), None)
        self.timeerr = None
        
        self._update_settings(devicesettings, generalsettings)

        self.rawdata = [None] * self.dcb_length

    def _update_settings(self, settings, generalsettings):
        """Check settings and update if needed."""

        if not generalsettings is None:
            for name, value in generalsettings.iteritems():
                setattr(self, '_' + name, value)

        for name, value in settings.iteritems():
            setattr(self, '_' + name, value)

        self.address = self._address #make externally available
        del self._address
        self.long_name = self._long_name #make externally available
        del self._long_name

        if self._expected_prog_mode == PROG_MODE_DAY:
            self.heat_schedule = SchedulerDayHeat()
            if self.is_hot_water():
                self.water_schedule = SchedulerDayWater()
        elif self._expected_prog_mode == PROG_MODE_WEEK:
            self.heat_schedule = SchedulerWeekHeat()
            if self.is_hot_water():
                self.water_schedule = SchedulerWeekWater()
        else:
            raise ValueError("Unknown program mode")

        self._expected_prog_mode_number = PROG_MODES[self._expected_prog_mode]

        self._fieldranges = FIELDRANGES[self._expected_model][self._expected_prog_mode]

        if self._expected_model == 'prt_e_model':
            self.dcb_map = PRTEmap[self._expected_prog_mode]
        elif self._expected_model == 'prt_hw_model':
            self.dcb_map = PRTHWmap[self._expected_prog_mode]
        elif self._expected_model == False:
            self.dcb_map = STRAIGHTmap
        else:
            raise ValueError("Unknown model %s"%self._expected_model)

        self._build_dcb_tables()

        self._expected_model_number = DEVICE_MODELS[self._expected_model]

        if self.dcb_map[0][1] != DCB_INVALID:
            self.dcb_length = self.dcb_map[0][0] - self.dcb_map[0][1] + 1
        elif self.dcb_map[1][1] != DCB_INVALID:
            self.dcb_length = self.dcb_map[1][0] - self.dcb_map[1][1] + 1
        else:
            raise ValueError("DCB map length not found")

        self.fullreadtime = self._estimate_read_time(self.dcb_length)
        
    def _get_dcb_address(self, uniqueaddress):
        """get the DCB address for a controller from the unique address"""
        return self._uniquetodcb[uniqueaddress]
    
    def _buildfieldtables(self):
        """build dict to map field name to index"""
        self._fieldnametonum = {}
        for key, data in enumerate(fields):
            fieldname = data[FIELD_NAME]
            self._fieldnametonum[fieldname] = key
                
    def _build_dcb_tables(self):
        """build list to map unique to dcb address and list of valid fields """
        #build a forward lookup table for the DCB values from unique address
        self._uniquetodcb = range(MAX_UNIQUE_ADDRESS+1)
        for uniquemax, offsetsel in self.dcb_map:
            self._uniquetodcb[0:uniquemax + 1] = [x - offsetsel for x in range(uniquemax + 1)] if not offsetsel is DCB_INVALID else [DCB_INVALID] * (uniquemax + 1)
        
        #build list of valid fields for this device
        self._fieldsvalid = [False] * len(fields)
        for first, last in self._fieldranges:
            self._fieldsvalid[self._fieldnametonum[first]: self._fieldnametonum[last] + 1] = [True] * (self._fieldnametonum[last] - self._fieldnametonum[first] + 1)
        #self._fullDCB = sum(x is not None for x in self._uniquetodcb))
        logging.debug("C%i Fieldsvalid %s"%(self.address, ','.join(str(int(x)) for x in self._fieldsvalid)))
    
    def _check_data_age(self, fieldnames, maxagein=None):
        """Check field data age is not more than maxage (in seconds)
        fieldnames can be list or string
        
        maxage = None, use the default from fields
        maxage = -1, only check if present
        maxage >=0, use maxage (0 is effectively always False)
        return False if old, True if recent"""
        if len(fieldnames) == 0:
            raise ValueError("Must list at least one field")
        
        if not isinstance(fieldnames, list):
            fieldnames = [fieldnames]
        
        for fieldname in fieldnames:
            if not self._check_data_present(fieldname):
                return False
            elif maxagein == -1: #only check present
                return True
            elif maxagein == None: #if none use field defaults
                maxage = fields[self._fieldnametonum[fieldname]][FIELD_MAX_AGE]
            else:
                maxage = maxagein
            #now check time
            if time.time() - self.datareadtime[fieldname] > maxage:
                logging.debug("C%i data item %s too old"%(self.address, fieldname))
                return False
        return True
        
    def _check_data_present(self, *fieldnames):
        """Check field(s) has data"""
        if len(fieldnames) == 0:
            raise ValueError("Must list at least one field")

        for fieldname in fieldnames:
            if self.datareadtime[fieldname] == None:
                logging.debug("C%i data item %s not available"%(self.address, fieldname))
                return False
        return True
    
    ## Basic reading and getting functions
    
    def read_all(self):
        """Returns all the field values having got them from the device"""
        try:
            self.rawdata = self._adaptor.read_all_from_device(self.address, self._protocol, self.dcb_length)
        except serial.SerialException as err:

            logging.warn("C%i Read all failed, Serial Port error %s"%(self.address, str(err)))
            raise
        else:
            logging.info("C%i Read all"%(self.address))

            self.lastreadtime = time.time()
            self._procpayload(self.rawdata)
            return self.rawdata

    def read_field(self, fieldname, maxage=None):
        """Returns a fields value, gets from the device if to old"""
        #return field value
        #get field from network if
        # maxage = None, older than the default from fields
        # maxage = -1, not read before
        # maxage >=0, older than maxage
        # maxage = 0, always
        if maxage == 0 or not self._check_data_age(fieldname, maxage):
            if self._autoreadall is True:
                self.get_field_range(fieldname)
            else:
                raise ValueError("Need to read %s first"%fieldname)
        return self.data[fieldname]
    
    def read_fields(self, fieldnames, maxage=None):
        """Returns a list of field values, gets from the device if any are to old"""
        #find which fields need getting because to old
        if not isinstance(fieldnames, list):
            raise TypeError("fieldnames must be a list")
        
        fieldids = [self._fieldnametonum[fieldname] for fieldname in fieldnames if self._fieldsvalid[self._fieldnametonum[fieldname]] and (maxage == 0 or not self._check_data_age(fieldname, maxage))]
        
        fieldids = list(set(fieldids)) #remove duplicates, ordering doesn't matter
        
        if len(fieldids) > 0 and self._autoreadall is True:
            self._get_fields(fieldids)
        elif len(fieldids) > 0:
            raise ValueError("Need to read fields first")
        return [self.data[fieldname] for fieldname in fieldnames]
    
    def get_variables(self):
        """Gets setroomtemp to hotwaterdemand fields from device"""
        self.get_field_range('setroomtemp', 'hotwaterdemand')
        
    def get_temps_and_demand(self):
        """Gets remoteairtemp to hotwaterdemand fields from device"""
        self.get_field_range('remoteairtemp', 'hotwaterdemand')
    
    def get_field_range(self, firstfieldname, lastfieldname=None):
        """gets fieldrange from device
        
        safe for blocks crossing gaps in dcb"""
        if lastfieldname == None:
            lastfieldname = firstfieldname

        blockstoread = self._get_field_blocks_from_range(firstfieldname, lastfieldname)
        logging.debug(blockstoread)
        estimatedreadtime = self._estimate_blocks_read_time(blockstoread)
        
        if estimatedreadtime < self.fullreadtime - 0.02: #if to close to full read time, then read all
            try:
                for firstfieldid, lastfieldid, blocklength in blockstoread:
                    logging.debug("C%i Reading ui %i to %i len %i, proc %s to %s"%(self.address, fields[firstfieldid][FIELD_ADD], fields[lastfieldid][FIELD_ADD], blocklength, fields[firstfieldid][FIELD_NAME], fields[lastfieldid][FIELD_NAME]))
                    rawdata = self._adaptor.read_from_device(self.address, self._protocol, fields[firstfieldid][FIELD_ADD], blocklength)
                    self.lastreadtime = time.time()
                    self._procpartpayload(rawdata, fields[firstfieldid][FIELD_NAME], fields[lastfieldid][FIELD_NAME])
            except serial.SerialException as err:
                logging.warn("C%i Read failed of fields %s to %s, Serial Port error %s"%(self.address, firstfieldname.ljust(FIELD_NAME_LENGTH), lastfieldname.ljust(FIELD_NAME_LENGTH), str(err)))
                raise
            else:
                logging.info("C%i Read fields %s to %s, in %i blocks"%(self.address, firstfieldname.ljust(FIELD_NAME_LENGTH), lastfieldname.ljust(FIELD_NAME_LENGTH), len(blockstoread)))
        else:
            logging.debug("C%i Read fields %s to %s by read_all, %0.3f %0.3f"%(self.address, firstfieldname.ljust(FIELD_NAME_LENGTH), lastfieldname.ljust(FIELD_NAME_LENGTH), estimatedreadtime, self.fullreadtime))
            self.read_all()

    def _get_fields(self, fieldids):
        """gets fields from device
        
        safe for blocks crossing gaps in dcb"""
        
        blockstoread = self._get_field_blocks_from_id_list(fieldids)
        logging.debug(blockstoread)
        estimatedreadtime = self._estimate_blocks_read_time(blockstoread)
        
        if estimatedreadtime < self.fullreadtime - 0.02: #if to close to full read time, then read all
            try:
                for firstfieldid, lastfieldid, blocklength in blockstoread:
                    logging.debug("C%i Reading ui %i to %i len %i, proc %s to %s"%(self.address, fields[firstfieldid][FIELD_ADD], fields[lastfieldid][FIELD_ADD], blocklength, fields[firstfieldid][FIELD_NAME], fields[lastfieldid][FIELD_NAME]))
                    rawdata = self._adaptor.read_from_device(self.address, self._protocol, fields[firstfieldid][FIELD_ADD], blocklength)
                    self.lastreadtime = time.time()
                    self._procpartpayload(rawdata, fields[firstfieldid][FIELD_NAME], fields[lastfieldid][FIELD_NAME])
            except serial.SerialException as err:
                logging.warn("C%i Read failed of fields %s, Serial Port error %s"%(self.address, ', '.join(fields[id][FIELD_NAME] for id in fieldids), str(err)))
                raise
            else:
                logging.info("C%i Read fields %s in %i blocks"%(self.address, ', '.join(fields[id][FIELD_NAME] for id in fieldids), len(blockstoread)))
                    
        else:
            logging.debug("C%i Read fields %s by read_all, %0.3f %0.3f"%(self.address, ', '.join(fields[id][FIELD_NAME] for id in fieldids), estimatedreadtime, self.fullreadtime))
            self.read_all()
                
        #data can only be requested from the controller in contiguous blocks
        #functions takes a first and last field and separates out the individual blocks available for the controller type
        #return, fieldstart, fieldend, length of read in bytes
    def _get_field_blocks_from_range(self, firstfieldname, lastfieldname):
        """Takes range of fieldnames and returns field blocks"""
        firstfieldid = self._fieldnametonum[firstfieldname]
        lastfieldid = self._fieldnametonum[lastfieldname]
        return self._get_field_blocks_from_id_range(firstfieldid, lastfieldid)
        
    def _get_field_blocks_from_id_range(self, firstfieldid, lastfieldid):
        """Takes range of fieldids and returns field blocks
        
        Splits by invalid fields"""
        blocks = []
        previousfieldvalid = False

        for fieldnum, fieldvalid in enumerate(self._fieldsvalid[firstfieldid:lastfieldid + 1], firstfieldid):
            if previousfieldvalid is False and not fieldvalid is False:
                start = fieldnum
            elif not previousfieldvalid is False and fieldvalid is False:
                blocks.append([start, fieldnum - 1, fields[fieldnum - 1][FIELD_ADD] + fields[fieldnum - 1][FIELD_LEN] - fields[start][FIELD_ADD]])
            
            previousfieldvalid = fieldvalid

        if not previousfieldvalid is False:
            blocks.append([start, lastfieldid, fields[lastfieldid][FIELD_ADD] + fields[lastfieldid][FIELD_LEN] - fields[start][FIELD_ADD]])
        return blocks
    
    def _get_field_blocks_from_id_list(self, fieldids):
        """Takes range of fieldids and returns field blocks
        
        Splits by invalid fields. Uses timing to determine the optimum blocking"""
        #find blocks between lowest and highest field
        fieldblocks = self._get_field_blocks_from_id_range(min(fieldids), max(fieldids))
        
        readblocks = []
        for block in fieldblocks:
            #find fields in that block
            inblock = [id for id in fieldids if block[0] <= id <= block[1]]
            if len(inblock) > 0:
                #if single read is shorter than individual
                readlen = fields[max(inblock)][FIELD_LEN] + fields[max(inblock)][FIELD_ADD] - fields[min(inblock)][FIELD_ADD]
                if self._estimate_read_time(readlen) < sum([self._estimate_read_time(fields[id][FIELD_LEN]) for id in inblock]):
                    readblocks.append([min(inblock), max(inblock), readlen])
                else:
                    for ids in inblock:
                        readblocks.append([ids, ids, fields[ids][FIELD_LEN]])
        return readblocks
    
    def _estimate_blocks_read_time(self, blocks):
        """estimates read time for a set of blocks, including the COM_BUS_RESET_TIME between blocks
        
        excludes the COM_BUS_RESET_TIME before first block"""
        readtimes = [self._estimate_read_time(x[2]) for x in blocks]
        return sum(readtimes) + self._adaptor.min_time_between_reads() * (len(blocks) - 1)
    
    @staticmethod
    def _estimate_read_time(length):
        """"estimates the read time for a call to read_from_device without COM_BUS_RESET_TIME
        
        based on empirical measurements of one prt_hw_model and 5 prt_e_model"""
        return length * 0.002075 + 0.070727
    
    def _procfield(self, data, fieldinfo, sentpayload=False):
        """Process data for a single field storing in relevant.
        
        Converts from bytes to integers/floats
        Checks the validity"""
        fieldname = fieldinfo[FIELD_NAME]
        length = fieldinfo[FIELD_LEN]
        factor = fieldinfo[FIELD_DIV]
        fieldrange = fieldinfo[FIELD_RANGE]
        #logging.debug("Processing %s %s"%(fieldinfo[FIELD_NAME],', '.join(str(x) for x in data)))
        if sentpayload: #some mapping required if was a sent payload
            data = self._map_write_to_read_payload(data, fieldname)
        
        if data is None:
            value = None
        elif length == 1:
            value = data[0]/factor
        elif length == 2:
            val_high = data[0]
            val_low = data[1]
            value = 1.0*(val_high*256 + val_low)/factor #force float, although always returns integer temps.
        elif length == 4:
            value = data
        elif length == 12:
            self.heat_schedule.set_raw(fieldname, data)
            value = data
        elif length == 16:
            self.water_schedule.set_raw(fieldname, data)
            value = data
        else:
            raise ValueError("_procpayload can't process field length")
    
        if len(fieldrange) == 2 and isinstance(fieldrange[0], (int, long)) and isinstance(fieldrange[1], (int, long)) and not value is None:
            if value < fieldrange[0] or value > fieldrange[1]:
                raise HeatmiserResponseError("Field value %i outside expected range"%value)
        
        if fieldname == 'DCBlen' and value != self.dcb_length:
            raise HeatmiserResponseError('DCBlengh is unexpected')
        
        if fieldname == 'model' and value != self._expected_model_number:
            raise HeatmiserResponseError('Model is unexpected')
        
        if fieldname == 'programmode' and value != self._expected_prog_mode_number:
            raise HeatmiserResponseError('Programme mode is unexpected')
        
        if fieldname == 'version' and self._expected_model != 'prt_hw_model':
            value = data[0] & 0x7f
            self.floorlimiting = data[0] >> 7
            self.data['floorlimiting'] = self.floorlimiting
        
        self.data[fieldname] = value
        setattr(self, fieldname, value)
        self.datareadtime[fieldname] = self.lastreadtime if not value is None else None #unless sent payload and don't know the read value
        
        if fieldname == 'currenttime':
            self._checkcontrollertime()
        
        ###todo, add range validation for other lengths

    @staticmethod
    def _map_write_to_read_payload(data, fieldname):
        """Maps written payload to equvialent read payload"""
        #returns None if unknown result.
        #handle odd effect on WRITE_hotwaterdemand_PROG
        if fieldname == 'hotwaterdemand':
            if data[0] == WRITE_HOTWATERDEMAND_PROG: #returned to program so outcome is unknown
                return None
            elif data[0] == WRITE_HOTWATERDEMAND_OVER_OFF: #if overridden off store the off read value
                return [READ_HOTWATERDEMAND_OFF]
            else:
                return data
        else:
            return data

    def _procpartpayload(self, rawdata, firstfieldname, lastfieldname, sentpayload=False):
        """Wraps procpayload by converting fieldnames to fieldids"""
        #rawdata must be a list
        #converts field names to unique addresses to allow process of shortened raw data
        logging.debug("C%i Processing Payload from field %s to %s"%(self.address, firstfieldname, lastfieldname))
        firstfieldid = self._fieldnametonum[firstfieldname]
        lastfieldid = self._fieldnametonum[lastfieldname]
        self._procpayload(rawdata, firstfieldid, lastfieldid, sentpayload)
        
    def _procpayload(self, rawdata, firstfieldid=0, lastfieldid=len(fields), sentpayload=False):
        """Split payload with field information and processes each field"""
        logging.debug("C%i Processing Payload from field %i to %i"%(self.address, firstfieldid, lastfieldid))

        fullfirstdcbadd = self._get_dcb_address(fields[firstfieldid][FIELD_ADD])
        
        for fieldinfo in fields[firstfieldid:lastfieldid + 1]:
            uniqueaddress = fieldinfo[FIELD_ADD]
            
            length = fieldinfo[FIELD_LEN]
            dcbadd = self._get_dcb_address(uniqueaddress)

            if dcbadd == DCB_INVALID:
                setattr(self, fieldinfo[FIELD_NAME], None)
                self.data[fieldinfo[FIELD_NAME]] = None
            else:
                dcbadd -= fullfirstdcbadd #adjust for the start of the request
                
                try:
                    self._procfield(rawdata[dcbadd:dcbadd+length], fieldinfo, sentpayload)
                except HeatmiserResponseError as err:
                    logging.warn("C%i Field %s process failed due to %s"%(self.address, fieldinfo[FIELD_NAME], str(err)))

        self.rawdata[fullfirstdcbadd:fullfirstdcbadd+len(rawdata)] = rawdata

    def _checkcontrollertime(self):
        """run check of device time against local read time, and try to fix if _autocorrectime"""
        try:
            self._comparecontrollertime()
        except HeatmiserControllerTimeError:
            if self._autocorrectime is True:
                self.set_time()
            else:
                raise
    
    def _comparecontrollertime(self):
        """Compare device and local time difference against threshold"""
        # Now do same sanity checking
        # Check the time is within range
        # currentday is numbered 1-7 for M-S
        # localday (python) is numbered 0-6 for Sun-Sat
        
        if not self._check_data_present('currenttime'):
            raise HeatmiserResponseError("Time not read before check")

        localtimearray = self._localtimearray(self.datareadtime['currenttime']) #time that time field was read
        localweeksecs = self._weeksecs(localtimearray)
        remoteweeksecs = self._weeksecs(self.data['currenttime'])
        directdifference = abs(localweeksecs - remoteweeksecs)
        wrappeddifference = abs(self.DAYSECS * 7 - directdifference) #compute the difference on rollover
        self.timeerr = min(directdifference, wrappeddifference)
        logging.debug("Local time %i, remote time %i, error %i"%(localweeksecs, remoteweeksecs, self.timeerr))

        if self.timeerr > self.DAYSECS:
            raise HeatmiserControllerTimeError("C%2d Incorrect day : local is %s, sensor is %s" % (self.address, localtimearray[CURRENT_TIME_DAY], self.data['currenttime'][CURRENT_TIME_DAY]))

        if self.timeerr > TIME_ERR_LIMIT:
            raise HeatmiserControllerTimeError("C%2d Time Error %d greater than %d: local is %s, sensor is %s" % (self.address, self.timeerr, TIME_ERR_LIMIT, localweeksecs, remoteweeksecs))

    @staticmethod
    def _localtimearray(timenow=time.time()):
        """creates an array in heatmiser format for local time. Day 1-7, 1=Monday"""
        #input time.time() (not local)
        localtimenow = time.localtime(timenow)
        nowday = localtimenow.tm_wday + 1 #python tm_wday, range [0, 6], Monday is 0
        nowsecs = min(localtimenow.tm_sec, 59) #python tm_sec range[0, 61]
        
        return [nowday, localtimenow.tm_hour, localtimenow.tm_min, nowsecs]
    
    DAYSECS = 86400
    HOURSECS = 3600
    MINSECS = 60
    def _weeksecs(self, localtimearray):
        """calculates the time from the start of the week in seconds from a heatmiser time array"""
        return (localtimearray[CURRENT_TIME_DAY] - 1) * self.DAYSECS + localtimearray[CURRENT_TIME_HOUR] * self.HOURSECS + localtimearray[CURRENT_TIME_MIN] * self.MINSECS + localtimearray[CURRENT_TIME_SEC]
    
    ## Basic set field functions
    
    def set_field(self, fieldname, payload):
        """Set a field (single member of fields) on a device to a state or payload. Defined for all known field lengths."""
        #Payload must not be list for field length 1 or 2
        fieldid = self._fieldnametonum[fieldname]
        if not self._fieldsvalid[fieldid]:
            raise IndexError('Field not valid for this device')
        fieldinfo = fields[fieldid]
        
        self._is_writable(fieldinfo)
        self._check_payload_values(payload, fieldinfo)
        payloadbytes = self._format_payload(payload, fieldinfo[FIELD_LEN])
        
        if not isinstance(payload, list): #adjust for the 
            payload = [payload]
        try:
            self._adaptor.write_to_device(self.address, self._protocol, fieldinfo[FIELD_ADD], fieldinfo[FIELD_LEN], payloadbytes)
        except serial.SerialException as err:
            logging.warn("C%i failed to set field %s to %s, due to %s"%(self.address, fieldname.ljust(FIELD_NAME_LENGTH), ', '.join(str(x) for x in payload), str(err)))
            raise
        else:
            logging.info("C%i set field %s to %s"%(self.address, fieldname.ljust(FIELD_NAME_LENGTH), ', '.join(str(x) for x in payload)))
        
        self.lastreadtime = time.time()
        self._procpartpayload(payloadbytes, fieldname, fieldname, True)
    
    @staticmethod
    def _format_payload(payload, field_len):
        """Convert fields to byte form for writting to device"""
        if field_len == 1:
            return [payload]
        elif field_len == 2:
            pay_lo = (payload & BYTEMASK)
            pay_hi = (payload >> 8) & BYTEMASK
            return [pay_lo, pay_hi]
        else:
            return payload
    
    @staticmethod
    def _is_writable(fieldinfo):
        """Checks if field is writable"""
        if len(fieldinfo) < FIELD_WRITE + 1 or fieldinfo[FIELD_WRITE] != 'W':
            #check that write is part of field info and is 'W'
            raise ValueError("set_field: field isn't writeable")
    
    def set_fields(self, fieldnames, payloads):
        """Set multiple fields on a device to a state or payload."""
        #It groups adjacent fields and issues multiple sets if required.
        #inputs must be mathcing length lists
        if not isinstance(fieldnames, list) or not isinstance(payloads, list):
            raise ValueError("fieldnames and payloads must be lists")
        if len(fieldnames) != len(payloads):
            raise IndexError("fields and payloads don't match")
        if len(fieldnames) != len(set(fieldnames)):
            raise ValueError("duplicated fieldnames")
        
        #Get field ids
        fieldids = [self._fieldnametonum[fieldname] for fieldname in fieldnames]
        outputdata = self._get_payload_blocks_from_list(fieldids, payloads)
        try:
            for unique_start_address, lengthbytes, payloadbytes, firstfieldname, lastfieldname in outputdata:
                logging.debug("C%i Setting ui %i len %i, proc %s to %s"%(self.address, unique_start_address, lengthbytes, firstfieldname, lastfieldname))
                self._adaptor.write_to_device(self.address, self._protocol, unique_start_address, lengthbytes, payloadbytes)
                self.lastreadtime = time.time()
                self._procpartpayload(payloadbytes, firstfieldname, lastfieldname, True)
        except serial.SerialException as err:
            logging.warn("C%i settings failed of fields %s, Serial Port error %s"%(self.address, ', '.join(fields[id][FIELD_NAME] for id in fieldids), str(err)))
            raise
        else:
            logging.info("C%i set fields %s in %i blocks"%(self.address, ', '.join(fields[id][FIELD_NAME] for id in fieldids), len(outputdata)))

    def _get_payload_blocks_from_list(self, fieldids, payloads):
        """Converts list of fields and payloads into groups of payload data"""
        #returns unique_start_address, lengthbytes, payloadbytes, firstfieldname, lastfieldname
        sortedfields = sorted(enumerate(fieldids), key=itemgetter(0))
        
        payloadscopy = copy.deepcopy(payloads) #force copy of payloads so doesn't get changed later.
        
        #Check field data and sort payloads
        sorteddata = []
        for orginalindex, fieldid in sortedfields:
            if self._fieldsvalid[fieldid]:
                fieldinfo = fields[fieldid]
                self._is_writable(fieldinfo)
                self._check_payload_values(payloadscopy[orginalindex], fieldinfo)
                sorteddata.append([fieldid, fieldinfo, payloadscopy[orginalindex]]) 
        
        #Groups and map to unique addresses
        outputdata = []
        previousfield = None
        for fieldid, fieldinfo, payload in sorteddata:
            if not previousfield is None and fieldid - previousfield == 1: #if follows previousfield
                outputdata[-1][1] += fieldinfo[FIELD_LEN]
                outputdata[-1][2].extend(self._format_payload(payload, fieldinfo[FIELD_LEN]))
                outputdata[-1][4] = fieldinfo[FIELD_NAME]
            else:
                #unique_start_address, bytelength, payloadbytes, firstfieldname, lastfieldname
                outputdata.append([fieldinfo[FIELD_ADD], fieldinfo[FIELD_LEN], self._format_payload(payload, fieldinfo[FIELD_LEN]), fieldinfo[FIELD_NAME], fieldinfo[FIELD_NAME]])
            previousfield = fieldid

        return outputdata
    
    @staticmethod
    def _check_payload_values(payload, fieldinfo):
        """check a single field payload matches field spec"""
        
        if fieldinfo[FIELD_LEN] in [1, 2] and not isinstance(payload, (int, long)):
            #one or two byte field, not single length payload
            raise TypeError("set_field: invalid requested value")
        elif fieldinfo[FIELD_LEN] > 2 and len(payload) != fieldinfo[FIELD_LEN]:
            #greater than two byte field, payload length must match field length
            raise ValueError("set_field: invalid payload length")

        #checks the payload matches the ranges if ranges are defined
        ranges = fieldinfo[FIELD_RANGE]
        if ranges != []:
            if isinstance(payload, (int, long)):
                if payload < ranges[0] or payload > ranges[1]:
                    raise ValueError("set_field: payload out of range")
            else:
                for i, item in enumerate(payload):
                    rangepair = ranges[i % len(ranges)]
                    if item < rangepair[0] or item > rangepair[1]:
                        raise ValueError("set_field: payload out of range")
    
    ## External functions for printing data
    def display_heating_schedule(self):
        """Prints heating schedule to stdout"""
        self.heat_schedule.display()
            
    def display_water_schedule(self):
        """Prints water schedule to stdout"""
        if not self.water_schedule is None:
            self.water_schedule.display()

    def print_target(self):
        """Returns text describing current heating state"""    
        current_state = self.read_temp_state()
        
        if current_state == self.TEMP_STATE_OFF:
            return "controller off without frost protection"
        elif current_state == self.TEMP_STATE_OFF_FROST:
            return "controller off"
        elif current_state == self.TEMP_STATE_HOLIDAY:
            return "controller on holiday for %i hours" % self.holidayhours
        elif current_state == self.TEMP_STATE_FROST:
            return "controller in frost mode"
        elif current_state == self.TEMP_STATE_HELD:
            return "temp held for %i mins at %i"%(self.tempholdmins, self.setroomtemp)
        elif current_state == self.TEMP_STATE_OVERRIDDEN:
            locatimenow = self._localtimearray()
            nexttarget = self.heat_schedule.get_next_schedule_item(locatimenow)
            return "temp overridden to %0.1f until %02d:%02d" % (self.setroomtemp, nexttarget[1], nexttarget[2])
        elif current_state == self.TEMP_STATE_PROGRAM:
            locatimenow = self._localtimearray()
            nexttarget = self.heat_schedule.get_next_schedule_item(locatimenow)
            return "temp set to %0.1f until %02d:%02d" % (self.setroomtemp, nexttarget[1], nexttarget[2])
    
    ## External functions for reading data

    def is_hot_water(self):
        """Does device manage hotwater?"""
        #returns True if stat is a model with hotwater control, False otherwise
        return self._expected_model == 'prt_hw_model'

    TEMP_STATE_OFF = 0    #thermostat display is off and frost protection disabled
    TEMP_STATE_OFF_FROST = 1 #thermostat display is off and frost protection enabled
    TEMP_STATE_FROST = 2 #frost protection enabled indefinitely
    TEMP_STATE_HOLIDAY = 3 #holiday mode, frost protection for a period
    TEMP_STATE_HELD = 4 #temperature held for a number of hours
    TEMP_STATE_OVERRIDDEN = 5 #temperature overridden until next program time
    TEMP_STATE_PROGRAM = 6 #following program
    
    def read_temp_state(self):
        """Returns the current temperature control state from off to following program"""
        self.read_fields(['mon_heat', 'tues_heat', 'wed_heat', 'thurs_heat', 'fri_heat', 'wday_heat', 'wend_heat'], -1)
        self.read_fields(['onoff', 'frostprot', 'holidayhours', 'runmode', 'tempholdmins', 'setroomtemp'])
        
        if self.onoff == WRITE_ONOFF_OFF and self.frostprot == READ_FROST_PROT_OFF:
            return self.TEMP_STATE_OFF
        elif self.onoff == WRITE_ONOFF_OFF and self.frostprot == READ_FROST_PROT_ON:
            return self.TEMP_STATE_OFF_FROST
        elif self.holidayhours != 0:
            return self.TEMP_STATE_HOLIDAY
        elif self.runmode == WRITE_RUNMODE_FROST:
            return self.TEMP_STATE_FROST
        elif self.tempholdmins != 0:
            return self.TEMP_STATE_HELD
        else:
        
            if not self._check_data_age(['currenttime'], MAX_AGE_MEDIUM):
                self.read_time()
            
            locatimenow = self._localtimearray()
            scheduletarget = self.heat_schedule.get_current_schedule_item(locatimenow)

            if scheduletarget[SCH_ENT_TEMP] != self.setroomtemp:
                return self.TEMP_STATE_OVERRIDDEN
            else:
                return self.TEMP_STATE_PROGRAM

    ### UNTESTED # last part about scheduletarget doesn't work
    def read_water_state(self):
        """Returns the current hot water control state from off to following program"""
        #does runmode affect hot water state?
        self.read_fields(['mon_water', 'tues_water', 'wed_water', 'thurs_water', 'fri_water', 'wday_water', 'wend_water'], -1)
        self.read_fields(['onoff', 'holidayhours', 'hotwaterdemand'])
        
        if self.onoff == WRITE_ONOFF_OFF:
            return self.TEMP_STATE_OFF
        elif self.holidayhours != 0:
            return self.TEMP_STATE_HOLIDAY
        else:
        
            if not self._check_data_age(['currenttime'], MAX_AGE_MEDIUM):
                self.read_time()
            
            locatimenow = self._localtimearray()
            scheduletarget = self.water_schedule.get_current_schedule_item(locatimenow)

            if scheduletarget[SCH_ENT_TEMP] != self.hotwaterdemand:
                return self.TEMP_STATE_OVERRIDDEN
            else:
                return self.TEMP_STATE_PROGRAM
                
    def read_air_sensor_type(self):
        """Reports airsensor type"""
        #1 local, 3 remote
        if not self._check_data_present('sensorsavaliable'):
            return False

        if self.sensorsavaliable == READ_SENSORS_AVALIABLE_INT_ONLY or self.sensorsavaliable == READ_SENSORS_AVALIABLE_INT_FLOOR:
            return 1
        elif self.sensorsavaliable == READ_SENSORS_AVALIABLE_EXT_ONLY or self.sensorsavaliable == READ_SENSORS_AVALIABLE_EXT_FLOOR:
            return 2
        else:
            return 0
            
    def read_air_temp(self):
        """Read the air temperature getting data from device if too old"""
        #if not read before read sensorsavaliable field
        self.read_field('sensorsavaliable', None)
        
        if self.sensorsavaliable == READ_SENSORS_AVALIABLE_INT_ONLY or self.sensorsavaliable == READ_SENSORS_AVALIABLE_INT_FLOOR:
            return self.read_field('airtemp', self._max_age_temp)
        elif self.sensorsavaliable == READ_SENSORS_AVALIABLE_EXT_ONLY or self.sensorsavaliable == READ_SENSORS_AVALIABLE_EXT_FLOOR:
            return self.read_field('remoteairtemp', self._max_age_temp)
        else:
            raise ValueError("sensorsavaliable field invalid")
    
    def read_raw_data(self, startfieldname=None, endfieldname=None):
        """Return subset of raw data"""
        if startfieldname == None or endfieldname == None:
            return self.rawdata
        else:
            return self.rawdata[self._get_dcb_address(uniadd[startfieldname][UNIADD_ADD]):self._get_dcb_address(uniadd[endfieldname][UNIADD_ADD])]
        
    def read_time(self, maxage=0):
        """Readtime, getting from device if required"""
        return self.read_field('currenttime', maxage)
        
    ## External functions for setting data

    def set_heating_schedule(self, day, schedule):
        """Set heating schedule for a single day"""
        padschedule = self.heat_schedule.pad_schedule(schedule)
        self.set_field(day, padschedule)
            
    def set_water_schedule(self, day, schedule):
        """Set water schedule for a single day"""
        padschedule = self.water_schedule.pad_schedule(schedule)
        if day == 'all':
            self.set_field('mon_water', padschedule)
            self.set_field('tues_water', padschedule)
            self.set_field('wed_water', padschedule)
            self.set_field('thurs_water', padschedule)
            self.set_field('fri_water', padschedule)
            self.set_field('sat_water', padschedule)
            self.set_field('sun_water', padschedule)
        else:
            self.set_field(day, padschedule)

    def set_time(self):
        """set time on device to match current localtime on server"""
        timenow = time.time() + 0.5 #allow a little time for any delay in setting
        return self.set_field('currenttime', self._localtimearray(timenow))

    #overriding

    def set_temp(self, temp):
        """sets the temperature demand overriding the program."""
        #Believe it returns at next prog change.
        if self.read_field('tempholdmins') == 0: #check hold temp not applied
            return self.set_field('setroomtemp', temp)
        else:
            logging.warn("%i address, temp hold applied so won't set temp"%(self.address))

    def release_temp(self):
        """release setTemp back to the program, but only if temp isn't held for a time (holdTemp)."""
        if self.read_field('tempholdmins') == 0: #check hold temp not applied
            return self.set_field('tempholdmins', 0)
        else:
            logging.warn("%i address, temp hold applied so won't remove set temp"%(self.address))

    def hold_temp(self, minutes, temp):
        """sets the temperature demand overrding the program for a set time."""
        #Believe it then returns to program.
        self.set_field('setroomtemp', temp)
        return self.set_field('tempholdmins', minutes)
        #didn't stay on if did minutes followed by temp.
        
    def release_hold_temp(self):
        """release setTemp or holdTemp back to the program."""
        return self.set_field('tempholdmins', 0)
        
    def set_holiday(self, hours):
        """sets holiday up for a defined number of hours."""
        return self.set_field('holidayhours', hours)
    
    def release_holiday(self):
        """cancels holiday mode"""
        return self.set_field('holidayhours', 0)

    #onoffs

    def set_on(self):
        """Switch stat on"""
        return self.set_field('onoff', WRITE_ONOFF_ON)
    def set_off(self):
        """Switch stat off"""
        return self.set_field('onoff', WRITE_ONOFF_OFF)
        
    def set_heat(self):
        """Switch stat to follow heat program"""
        return self.set_field('runmode', WRITE_RUNMODE_HEATING)
    def set_frost(self):
        """Switch stat to frost only"""
        return self.set_field('runmode', WRITE_RUNMODE_FROST)
        
    def set_lock(self):
        """Lock keypad"""
        return self.set_field('keylock', WRITE_KEYLOCK_ON)
    def set_unlock(self):
        """Unlock keypad"""
        return self.set_field('keylock', WRITE_KEYLOCK_OFF)
    
#other
#set floor limit

#create a controller that broadcasts or reads from multiple stats
class HeatmiserBroadcastDevice(HeatmiserDevice):
    """Broadcast device class for broadcast set functions and managing reading on all devices"""
    _controllerlist = ListWrapperClass()

    def __init__(self, network, long_name, controllerlist=None):
        self._controllerlist.list = controllerlist
        settings = {
            'address':BROADCAST_ADDR,
            'display_order': 0,
            'long_name': long_name,
            'protocol':DEFAULT_PROTOCOL,
            'expected_model':False,
            'expected_prog_mode':DEFAULT_PROG_MODE
            }
        super(HeatmiserBroadcastDevice, self).__init__(network, settings)
    
    #run read functions on all stats
    @run_function_on_all(_controllerlist)
    def read_field(self, fieldname, maxage=None):
        logging.info("All reading %s from %i controllers"%(fieldname, len(self._controllerlist.list)))
            
    @run_function_on_all(_controllerlist)
    def read_fields(self, fieldnames, maxage=None):
        logging.info("All reading %s from %i controllers"%(', '.join([fieldname for fieldname in fieldnames]), len(self._controllerlist.list)))
        
    @run_function_on_all(_controllerlist)
    def read_air_temp(self):
        pass
    
    @run_function_on_all(_controllerlist)
    def read_temp_state(self):
        pass
    
    @run_function_on_all(_controllerlist)
    def read_water_state(self):
        pass
    
    @run_function_on_all(_controllerlist)
    def read_air_sensor_type(self):
        pass
            
    @run_function_on_all(_controllerlist)
    def read_time(self, maxage=0):
        pass
    
    #run set functions which require a read on all stats
    @run_function_on_all(_controllerlist)
    def set_temp(self, temp):
        pass
    
    @run_function_on_all(_controllerlist)
    def release_temp(self):
        pass




"""Constants and field definitions for Heatmiser protocol"""

# Protocols
HMV2_ID = 2
HMV3_ID = 3
DEFAULT_PROTOCOL = HMV3_ID

BYTEMASK = 0xff

DONT_CARE_LENGTH = 1
TIME_ERR_LIMIT = 10

#
# HM Version 3 Magic Numbers
#

# Master must be in range [0x81, 0xa0] = [129, 160]
MASTER_ADDR_MIN = 0x81
MASTER_ADDR_MAX = 0xa0
SLAVE_ADDR_MIN = 1
SLAVE_ADDR_MAX = 32

MAX_FRAME_RESP_LENGTH = 159 # NB max return is 75 in 5/2 mode or 159 in 7day mode
MIN_FRAME_RESP_LENGTH = 7
FRAME_WRITE_RESP_LENGTH = 7
MIN_FRAME_READ_RESP_LENGTH = 11
MIN_FRAME_SEND_LENGTH = 10
MAX_PAYLOAD_SEND_LENGTH = 100
CRC_LENGTH = 2

# Define magic numbers used in messages
FUNC_READ = 0
FUNC_WRITE = 1

BROADCAST_ADDR = 0xff
RW_LENGTH_ALL = 0xffff
DCB_START = 0x00

# Define frame send contents
FS_DEST_ADDR = 0
FS_LEN = 1
FS_SOURCE_ADDR = 2
FS_FUNC_CODE = 3

# Define frame response contents
FR_DEST_ADDR = 0
FR_LEN_LOW = 1
FR_LEN_HIGH = 2
FR_SOURCE_ADDR = 3
FR_FUNC_CODE = 4
FR_START_LOW = 5 # this field and following on present if function code read
FR_START_HIGH = 6
FR_CONT_LEN_LOW = 7
FR_CONT_LEN_HIGH = 8
FR_CONTENTS = 9

MAX_AGE_LONG = 86400
MAX_AGE_MEDIUM = 3600
MAX_AGE_SHORT = 65
MAX_AGE_USHORT = 11

FIELD_NAME_LENGTH = 13
MAX_UNIQUE_ADDRESS = 298
#name,  unique_address, length, divisor,  valid range,  writeable]
FIELD_NAME = 0
FIELD_ADD = 1
FIELD_LEN = 2
FIELD_DIV = 3
FIELD_RANGE = 4
FIELD_MAX_AGE = 5
FIELD_WRITE = 6
fields = [
['DCBlen', 0, 2, 1, [], MAX_AGE_LONG],
['vendor', 2, 1, 1, [0, 1], MAX_AGE_LONG],  #00 heatmiser,  01 OEM
['version', 3, 1, 1, [], MAX_AGE_LONG],
['model', 4, 1, 1, [0, 5], MAX_AGE_LONG],  # DT/DT-E/PRT/PRT-E 00/01/02/03
['tempformat', 5, 1, 1, [0, 1], MAX_AGE_LONG],  # 00 C,  01 F
['switchdiff', 6, 1, 1, [1, 3], MAX_AGE_LONG],
['frostprot', 7, 1, 1, [0, 1], MAX_AGE_LONG],  #0=enable frost prot when display off,  (opposite in protocol manual,  but tested and user guide is correct)  (default should be enabled)
['caloffset', 8, 2, 1, [], MAX_AGE_LONG],
['outputdelay', 10, 1, 1, [0, 15], MAX_AGE_LONG],  # minutes (to prevent rapid switching)
['address', 11, 1, 1, [SLAVE_ADDR_MIN, SLAVE_ADDR_MAX], MAX_AGE_LONG],
['updwnkeylimit', 12, 1, 1, [0, 10], MAX_AGE_LONG],   #limits use of up and down keys
['sensorsavaliable', 13, 1, 1, [0, 4], MAX_AGE_LONG],  #00 built in only,  01 remote air only,  02 floor only,  03 built in + floor,  04 remote + floor
['optimstart', 14, 1, 1, [0, 3], MAX_AGE_LONG],  # 0 to 3 hours,  default 0
['rateofchange', 15, 1, 1, [], MAX_AGE_LONG],  #number of minutes per degree to raise the temperature,  default 20. Applies to the Wake and Return comfort levels (1st and 3rd)
['programmode', 16, 1, 1, [0, 1], MAX_AGE_LONG],  #0=5/2,  1= 7day
['frosttemp', 17, 1, 1, [7, 17], MAX_AGE_LONG, 'W'],  #default is 12,  frost protection temperature
['setroomtemp', 18, 1, 1, [5, 35], MAX_AGE_USHORT, 'W'],
['floormaxlimit', 19, 1, 1, [20, 45], MAX_AGE_LONG, 'W'],
['floormaxlimitenable', 20, 1, 1, [0, 1], MAX_AGE_LONG],  #1=enable
['onoff', 21, 1, 1, [0, 1], MAX_AGE_SHORT, 'W'],  #1 = on
['keylock', 22, 1, 1, [0, 1], MAX_AGE_SHORT, 'W'],  #1 = on
['runmode', 23, 1, 1, [0, 1], MAX_AGE_SHORT, 'W'],   #0 = heating mode,  1 = frost protection mode
['holidayhours', 24, 2, 1, [0, 720], MAX_AGE_SHORT, 'W'],  #range guessed and tested,  setting to 0 cancels hold and puts back to program 
['unknown', 26, 6, 1, [], MAX_AGE_LONG],  # gap from 26 to 31
['tempholdmins', 32, 2, 1, [0, 5760], MAX_AGE_SHORT, 'W'],  #range guessed and tested,  setting to 0 cancels hold and puts setroomtemp back to program
['remoteairtemp', 34, 2, 10, [], MAX_AGE_USHORT],  #ffff if no sensor
['floortemp', 36, 2, 10, [], MAX_AGE_USHORT],  #ffff if no sensor
['airtemp', 38, 2, 10, [], MAX_AGE_USHORT],  #ffff if no sensor
['errorcode', 40, 1, 1, [0, 3], MAX_AGE_SHORT],  # 0 is no error # errors,  0 built in,  1,  floor,  2 remote
['heatingdemand', 41, 1, 1, [0, 1], MAX_AGE_USHORT],  #0 none,  1 heating currently
['hotwaterdemand', 42, 1, 1, [0, 2], MAX_AGE_USHORT, 'W'],  # read [0=off, 1=on],  write [0=as prog, 1=override on, 2=overide off]
['currenttime', 43, 4, 1, [[1, 7], [0, 23], [0, 59], [0, 59]], MAX_AGE_USHORT, 'W'],  #day (Mon - Sun),  hour,  min,  sec.
#5/2 progamming #if hour = 24 entry not used
['wday_heat', 47, 12, 1, [[0, 24], [0, 59], [5, 35]], MAX_AGE_MEDIUM, "W"],  #hour,  min,  temp  (should minutes be only 0 and 30?)
['wend_heat', 59, 12, 1, [[0, 24], [0, 59], [5, 35]], MAX_AGE_MEDIUM, "W"],
['wday_water', 71, 16, 1, [[0, 24], [0, 59]], MAX_AGE_MEDIUM, "W"],  # pairs,  on then off repeated,  hour,  min
['wend_water', 87, 16, 1, [[0, 24], [0, 59]], MAX_AGE_MEDIUM, "W"],
#7day progamming
['mon_heat', 103, 12, 1, [[0, 24], [0, 59], [5, 35]], MAX_AGE_MEDIUM, "W"],
['tues_heat', 115, 12, 1, [[0, 24], [0, 59], [5, 35]], MAX_AGE_MEDIUM, "W"],
['wed_heat', 127, 12, 1, [[0, 24], [0, 59], [5, 35]], MAX_AGE_MEDIUM, "W"],
['thurs_heat', 139, 12, 1, [[0, 24], [0, 59], [5, 35]], MAX_AGE_MEDIUM, "W"],
['fri_heat', 151, 12, 1, [[0, 24], [0, 59], [5, 35]], MAX_AGE_MEDIUM, "W"],
['sat_heat', 163, 12, 1, [[0, 24], [0, 59], [5, 35]], MAX_AGE_MEDIUM, "W"],
['sun_heat', 175, 12, 1, [[0, 24], [0, 59], [5, 35]], MAX_AGE_MEDIUM, "W"],
['mon_water', 187, 16, 1, [[0, 24], [0, 59]], MAX_AGE_MEDIUM, "W"],
['tues_water', 203, 16, 1, [[0, 24], [0, 59]], MAX_AGE_MEDIUM, "W"],
['wed_water', 219, 16, 1, [[0, 24], [0, 59]], MAX_AGE_MEDIUM, "W"],
['thurs_water', 235, 16, 1, [[0, 24], [0, 59]], MAX_AGE_MEDIUM, "W"],
['fri_water', 251, 16, 1, [[0, 24], [0, 59]], MAX_AGE_MEDIUM, "W"],
['sat_water', 267, 16, 1, [[0, 24], [0, 59]], MAX_AGE_MEDIUM, "W"],
['sun_water', 283, 16, 1, [[0, 24], [0, 59]], MAX_AGE_MEDIUM, "W"]
]

CURRENT_TIME_DAY = 0
CURRENT_TIME_HOUR = 1
CURRENT_TIME_MIN = 2
CURRENT_TIME_SEC = 3

READ_FROST_PROT_ON = 0
READ_FROST_PROT_OFF = 1

READ_SENSORS_AVALIABLE_INT_ONLY = 00
READ_SENSORS_AVALIABLE_EXT_ONLY = 01
READ_SENSORS_AVALIABLE_FLOOR_ONLY = 02
READ_SENSORS_AVALIABLE_INT_FLOOR = 03
READ_SENSORS_AVALIABLE_EXT_FLOOR = 04

READ_PROGRAM_MODE_5_2 = 0
READ_PROGRAM_MODE_7 = 1

WRITE_ONOFF_OFF = 0
WRITE_ONOFF_ON = 1
WRITE_KEYLOCK_OFF = 0
WRITE_KEYLOCK_ON = 1

WRITE_RUNMODE_HEATING = 0
WRITE_RUNMODE_FROST = 1

WRITE_HOTWATERDEMAND_PROG = 0
WRITE_HOTWATERDEMAND_OVER_ON = 1
WRITE_HOTWATERDEMAND_OVER_OFF = 2
READ_HOTWATERDEMAND_OFF = 0
READ_HOTWATERDEMAND_ON = 1

#these maps are used to translate the unique address which is the same for all controllers to the DCB address for a specific controller
PROG_MODE_WEEK = 'week'
PROG_MODE_DAY = 'day'
DEFAULT_PROG_MODE = PROG_MODE_DAY #allows broadcast to program both modes.
PROG_MODES = {'week':0, 'day':1}

#models
DEVICE_MODELS = {'prt_e_model': 3, 'prt_hw_model': 4, False: 0}
#PRT-E DCB map
DCB_INVALID = None
PRTEmap = {}
PRTEmap[PROG_MODE_WEEK] = list(reversed([(25, 0), (31, DCB_INVALID), (41, 6), (42, DCB_INVALID), (70, 7), (MAX_UNIQUE_ADDRESS, DCB_INVALID)]))
PRTEmap[PROG_MODE_DAY] = list(reversed([(25, 0), (31, DCB_INVALID), (41, 6), (42, DCB_INVALID), (70, 7), (102, DCB_INVALID), (186, 39), (MAX_UNIQUE_ADDRESS, DCB_INVALID)]))
PRTHWmap = {}
PRTHWmap[PROG_MODE_WEEK] = list(reversed([(25, 0), (31, DCB_INVALID), (102, 6), (MAX_UNIQUE_ADDRESS, DCB_INVALID)]))
PRTHWmap[PROG_MODE_DAY] = list(reversed([(25, 0), (31, DCB_INVALID), (MAX_UNIQUE_ADDRESS, 6)]))
STRAIGHTmap = list([(MAX_UNIQUE_ADDRESS, 0)])

#Fields for each stat type and mode,  list of ranges.
#PRT-E fields
FIELDRANGES = {}
FIELDRANGES['prt_e_model'] = {}
FIELDRANGES['prt_e_model'][PROG_MODE_WEEK] = [['DCBlen', 'holidayhours'], ['tempholdmins', 'heatingdemand'], ['currenttime', 'wend_heat']]
FIELDRANGES['prt_e_model'][PROG_MODE_DAY] = [['DCBlen', 'holidayhours'], ['tempholdmins', 'heatingdemand'], ['currenttime', 'wend_heat'], ['mon_heat', 'sun_heat']]
#PRT-HW fields
FIELDRANGES['prt_hw_model'] = {}
FIELDRANGES['prt_hw_model'][PROG_MODE_WEEK] = [['DCBlen', 'holidayhours'], ['tempholdmins', 'wend_water']]
FIELDRANGES['prt_hw_model'][PROG_MODE_DAY] = [['DCBlen', 'holidayhours'], ['tempholdmins', 'sun_water']]
#Full map
FIELDRANGES[False] = {PROG_MODE_WEEK: [['DCBlen', 'sun_water']], PROG_MODE_DAY: [['DCBlen', 'sun_water']]}



"""Functions for creating and checking Heatmiser protocol frames"""
import logging

from hm_constants import *
from .exceptions import HeatmiserResponseError, HeatmiserResponseErrorCRC

### low level framing functions

def form_read_frame(destination, protocol, source, start, length):
    """Forms a read message payload, including CRC"""
    return form_frame(destination, protocol, source, FUNC_READ, start, length, [])

# TODO check master address is in legal range
def form_frame(destination, protocol, source, function, start, length, payload):
    """Forms a message payload, including CRC"""
    if protocol != HMV3_ID:
        raise ValueError("Protocol unknown")
    start_low = start & BYTEMASK
    start_high = (start >> 8) & BYTEMASK
    length_low = length & BYTEMASK
    length_high = (length >> 8) & BYTEMASK
    payload_length = len(payload)
    frame_length = MIN_FRAME_SEND_LENGTH
    if function == FUNC_WRITE:
        if length != payload_length:
            raise ValueError("Payload doesn't match length %s" % length)
        if length > MAX_PAYLOAD_SEND_LENGTH:
            raise ValueError("Payload to long %s" % length)
        frame_length = MIN_FRAME_SEND_LENGTH + payload_length
    msg = [destination, frame_length, source, function, start_low, start_high, length_low, length_high]
    if function == FUNC_WRITE:
        msg = msg + payload

    crc = crc16()
    msg = msg + crc.run(msg)
    return msg

def _check_frame_crc(protocol, data):
    """Takes frame with CRC and checks it is valid"""
    datalength = len(data)

    if protocol != HMV3_ID:
        raise ValueError("Protocol unknown")
    if datalength < 2:
        raise HeatmiserResponseError("No CRC")

    checksum = data[len(data)-2:]
    rxmsg = data[:len(data)-2]

    crc = crc16() # Initialises the CRC
    expectedchecksum = crc.run(rxmsg)
    if expectedchecksum != checksum:
        raise HeatmiserResponseErrorCRC("CRC is incorrect")

def _check_response_frame_length(protocol, data, expected_length):
    """Takes frame and checks length, must be a receive frame"""
    if protocol != HMV3_ID:
        raise ValueError("Protocol unknown")

    if len(data) < MIN_FRAME_RESP_LENGTH:
        raise HeatmiserResponseError("Response length too short: %s %s"%(len(data), MIN_FRAME_RESP_LENGTH))

    frame_len_l = data[FR_LEN_LOW]
    frame_len_h = data[FR_LEN_HIGH]
    frame_len = (frame_len_h << 8) | frame_len_l
    func_code = data[FR_FUNC_CODE]
    
    if len(data) != frame_len:
        raise HeatmiserResponseError("Frame length does not match header: %s %s" %(len(data), frame_len))

    if expected_length != RW_LENGTH_ALL and func_code == FUNC_READ and frame_len != MIN_FRAME_READ_RESP_LENGTH + expected_length:
        # Read response length is wrong
        raise HeatmiserResponseError("Response length %s not EXPECTED value %s + %s given read request" %(frame_len, MIN_FRAME_READ_RESP_LENGTH, expected_length))
    if func_code == FUNC_WRITE and frame_len != FRAME_WRITE_RESP_LENGTH:
        # Reply to Write is always 7 long
        raise HeatmiserResponseError("Response length %s not EXPECTED value %s given write request" %(frame_len, FRAME_WRITE_RESP_LENGTH))
            
def _check_response_frame_addresses(protocol, source, destination, data):
    """Takes frame and checks addresses are correct"""
    if protocol != HMV3_ID:
        raise ValueError("Protocol unknown")
        
    dest_addr = data[FR_DEST_ADDR]
    source_addr = data[FR_SOURCE_ADDR]

    if dest_addr < MASTER_ADDR_MIN or dest_addr > MASTER_ADDR_MAX:
        raise HeatmiserResponseError("Destination address out of valid range %i" % dest_addr)
    if dest_addr != destination:
        raise HeatmiserResponseError("Destination address incorrect %i" % dest_addr)
    if source_addr < SLAVE_ADDR_MIN or source_addr > SLAVE_ADDR_MAX:
        raise HeatmiserResponseError("Source address out of valid range %i" % source_addr)
    if source_addr != source:
        raise HeatmiserResponseError("Source address does not match %i" % source_addr)
            
def _check_response_frame_function(protocol, expected_function, data):
    """Takes frame and read or write bit is set correctly"""
    if protocol != HMV3_ID:
        raise ValueError("Protocol unknown")

    func_code = data[FR_FUNC_CODE]

    if func_code != FUNC_WRITE and func_code != FUNC_READ:
        raise HeatmiserResponseError("Unknown function    code: %i" %(func_code))
    if func_code != expected_function:
        raise HeatmiserResponseError("Function    code was not as expected: %i" %(func_code))
        
def verify_write_ack(protocol, source, destination, data):
    """Verifies message response to write is correct"""
    return verify_response(protocol, source, destination, FUNC_WRITE, DONT_CARE_LENGTH, data)
 
def verify_response(protocol, source, destination, expected_function, expected_length, data):
    """Verifies response frame appears legal"""
    try:
        # check CRC
        _check_frame_crc(protocol, data)
        # check length
        _check_response_frame_length(protocol, data, expected_length)
        # check addresses
        _check_response_frame_addresses(protocol, source, destination, data)
        # check function
        _check_response_frame_function(protocol, expected_function, data)
    except HeatmiserResponseError as err:
        logging.warning("C%s Invalid Response: %s: %s" %(source, str(err), data))
        raise
        
    ## missing check that it is valid for this type of controller. Use DCBUnique function not false.
    ## although if needed should be in devices and not in framing

# Believe this is known as CCITT (0xFFFF)
# This is the CRC function converted directly from the Heatmiser C code
# provided in their API
class crc16:
    """Computes CRC for Heatmiser Message"""
    LookupHigh = [
    0x00, 0x10, 0x20, 0x30, 0x40, 0x50, 0x60, 0x70,
    0x81, 0x91, 0xa1, 0xb1, 0xc1, 0xd1, 0xe1, 0xf1
    ]
    LookupLow = [
    0x00, 0x21, 0x42, 0x63, 0x84, 0xa5, 0xc6, 0xe7,
    0x08, 0x29, 0x4a, 0x6b, 0x8c, 0xad, 0xce, 0xef
    ]
    def __init__(self):
        self.high = BYTEMASK
        self.low = BYTEMASK

    def _update_4_bits(self, val):
        # Step one, extract the Most significant 4 bits of the CRC register
        #print "val is %d" %(val)
        t = self.high>>4
        #print "t is %d" %(t)

        # XOR in the Message Data into the extracted bits
        t = t^val
        #print "t is %d" %(t)

        # Shift the CRC Register left 4 bits
        self.high = (self.high << 4)|(self.low>>4)
        self.high = self.high & BYTEMASK # force char
        self.low = self.low <<4
        self.low = self.low & BYTEMASK # force char

        # Do the table lookups and XOR the result into the CRC tables
        #print "t for lookup is %d" %(t)
        self.high = self.high ^ self.LookupHigh[t]
        self.high = self.high & BYTEMASK # force char
        self.low = self.low ^ self.LookupLow[t]
        self.low = self.low & BYTEMASK # force char
        #print "high is %d Low is %d" %(self.high, self.low)

    def _crc16_update(self, val):
        self._update_4_bits(val>>4) # High nibble first
        self._update_4_bits(val & 0x0f) # Low nibble

    def run(self, message):
        """Calculates a CRC"""
        for c in message:
            #print c
            self._crc16_update(c)
        #print "CRC is Low %d High    %d" %(self.low, self.high)
        return [self.low, self.high]



#!/usr/bin/env python
"""Example of setting configurations on Heatmiser stats
Ian Horsley 2018
"""

import logging

from heatmisercontroller.logging_setup import initialize_logger
from heatmisercontroller.hm_constants import *
from heatmisercontroller.network import *

initialize_logger('logs', logging.INFO, True)

HMN = HeatmiserNetwork()

#HMN.set_temp("Kit", 25)
#HMN.Kit.holdTemp(30,21) #mins, temp
#HMN.Kit.releaseHoldTemp()

#HMN.set_temp("Cons", 25)
#HMN.release_temp("Cons")
#HMN.All.set_field('holidayhours', 96)
HMN.All.release_holiday()

HMN.Kit.set_field('hotwaterdemand', WRITE_HOTWATERDEMAND_PROG)

HMN.All.set_on()
#HMN.hmUpdateTime(2)
#HMN.controllerByName('B1').setOff()

HMN.All.set_field('runmode', WRITE_RUNMODE_HEATING)
HMN.Sit.set_field('runmode', WRITE_RUNMODE_FROST)

#HMN.B2.set_temp(24)
#HMN.Kit.set_temp(24)
#HMN.B1.set_temp(24)
#HMN.Cons.set_temp(24)
#HMN.All.set_temp(24)

HMN.Kit.release_temp()
HMN.B1.release_temp()
HMN.B2.release_temp()
HMN.Cons.release_temp()

DOWNINACTIVE = 20
DOWNACTIVE = 19
FROST = 12
DAYFROST = [7, 0, FROST]

WKDAY_ZEARLY = [5, 00, DOWNACTIVE, 8, 30, FROST, 15, 00, DOWNINACTIVE, 21, 00, FROST]
WKDAY_ZLATE = [7, 00, DOWNINACTIVE, 12, 00, FROST, 17, 00, DOWNINACTIVE, 21, 30, FROST]
WKDAY_ZTRAINING = [7, 00, DOWNACTIVE, 8, 30, FROST, 16, 30, DOWNINACTIVE, 21, 00, FROST]
WKEND_ZEARLY = [5, 00, DOWNINACTIVE, 21, 30, FROST]
WKEND_ZLATE = [7, 00, DOWNINACTIVE, 21, 30, FROST]
WKEND_ZOFF = [7, 00, DOWNINACTIVE, 21, 30, FROST]
HMN.Kit.set_heating_schedule('mon_heat', WKEND_ZEARLY)
HMN.Kit.set_heating_schedule('tues_heat', WKEND_ZEARLY)
HMN.Kit.set_heating_schedule('wed_heat', WKDAY_ZTRAINING)
HMN.Kit.set_heating_schedule('thurs_heat', WKEND_ZEARLY)
HMN.Kit.set_heating_schedule('fri_heat', WKEND_ZEARLY)
HMN.Kit.set_heating_schedule('sat_heat', WKEND_ZOFF)
HMN.Kit.set_heating_schedule('sun_heat', WKEND_ZOFF)

EVENINGWATER = [17, 30, 18, 0]
NOWATER = []
HMN.Kit.set_water_schedule('all', EVENINGWATER)

UPSLEEP = 16
UPAWAKE = 18
UPWKDAY_ZEARLY = [5, 0, UPAWAKE, 7, 00, FROST, 20, 30, UPAWAKE, 21, 30, UPSLEEP]
UPWKDAY_ZLATE = [7, 0, UPAWAKE, 8, 00, FROST, 21, 30, UPAWAKE, 22, 30, UPSLEEP]
UPWKDAY = [7, 0, UPAWAKE, 8, 30, FROST, 21, 00, UPAWAKE, 21, 30, UPSLEEP]
UPWKEND_ZEARLY = [5, 0, UPAWAKE, 8, 00, FROST, 20, 30, UPAWAKE, 21, 30, UPSLEEP]
UPWKEND_ZLATE = [7, 0, UPAWAKE, 9, 30, FROST, 21, 30, UPAWAKE, 22, 30, UPSLEEP]
UPWKEND = [7, 0, UPAWAKE, 9, 30, FROST, 21, 00, UPAWAKE, 22, 00, UPSLEEP]
DAYFROST = [7, 0, FROST]

HMN.B1.set_heating_schedule('mon_heat', DAYFROST)
HMN.B1.set_heating_schedule('tues_heat', DAYFROST)
HMN.B1.set_heating_schedule('wed_heat', DAYFROST)
HMN.B1.set_heating_schedule('thurs_heat', DAYFROST)
HMN.B1.set_heating_schedule('fri_heat', DAYFROST)
HMN.B1.set_heating_schedule('sat_heat', DAYFROST)
HMN.B1.set_heating_schedule('sun_heat', DAYFROST)

HMN.B2.set_heating_schedule('mon_heat', DAYFROST)
HMN.B2.set_heating_schedule('tues_heat', DAYFROST)
HMN.B2.set_heating_schedule('wed_heat', DAYFROST)
HMN.B2.set_heating_schedule('thurs_heat', DAYFROST)
HMN.B2.set_heating_schedule('fri_heat', DAYFROST)
HMN.B2.set_heating_schedule('sat_heat', DAYFROST)
HMN.B2.set_heating_schedule('sun_heat', DAYFROST)

#HMN.hmSetFields('Kit', 'wday_heat', [7, 0, 19, 9, 30, 10, 17, 0, 19, 21, 30, 10])
#HMN.hmSetFields('B1', HMV3_ID,'wday_heat', [7, 0, 18, 8, 30, 10, 20, 30, 18, 22, 0, 16])
#HMN.hmSetFields('B2', HMV3_ID,'wday_heat', [7, 0, 19, 8, 30, 10, 20, 30, 19, 22, 0, 16])
HMN.get_controller_by_name('Cons').set_field('wday_heat', [9, 0, 12, 21, 30, 10, 24, 0, 5, 24, 0, 5])

#HMN.hmSetFields('Kit', 'wend_heat', [7, 0, 19, 21, 30, 10, 24, 0, 5, 24, 0, 5])
#HMN.hmSetFields('B1', HMV3_ID,'wend_heat', [7, 0, 18, 9, 30, 10, 20, 30, 18, 22, 0, 16])
#HMN.hmSetFields('B2', HMV3_ID,'wend_heat', [7, 0, 19, 9, 30, 10, 20, 30, 19, 22, 0, 16])
HMN.get_controller_by_name('Cons').set_field('wend_heat', [9, 0, 12, 21, 30, 10, 24, 0, 5, 24, 0, 5])

HMN.Sit.set_heating_schedule('wday_heat', DAYFROST)
HMN.Sit.set_heating_schedule('wend_heat', DAYFROST)

#HMN.hmSetFields("Kit", 'wday_water', [7, 0, 8, 0, 16, 0, 17, 0, 24, 0, 24, 0, 24, 0, 24, 0])
#HMN.hmSetFields("Kit", 'wday_water', [24, 0, 24, 0, 24, 0, 24, 0, 24, 0, 24, 0, 24, 0, 24, 0])
#HMN.hmSetFields("Kit", 'wend_water', [8, 0, 9, 0, 18, 0, 19, 0, 24, 0, 24, 0, 24, 0, 24, 0])

#HMN.hmSetField("Kit", 'frosttemp', 9)
#HMN.hmSetField("B1", 'frosttemp', 9)
#HMN.hmSetField("B2", 'frosttemp', 9)
#HMN.hmSetField("Cons", 'frosttemp', 9)





from setuptools import setup

def readme():
    with open('README.rst') as f:
        return f.read()

setup(name='heatmisercontroller',
      version='0.33',
      description='Python implementation of Heatmiser protocol for serial connected thermostats ',
      long_description=readme(),
      classifiers=[
        'Programming Language :: Python :: 2.7',
      ],
      url='https://github.com/ianhorsley/heatmisercontroller',
      author='Ian Horsley',
      #author_email='flyingcircus@example.com',
      license='GNU v3.0',
      packages=['heatmisercontroller'],
      install_requires=[
        'datetime',
        'logging',
        'pyserial',
        'configobj'
      ],
      test_suite="tests",
      scripts=[
        'bin/hm_get_example.py',
        'bin/hm_set_example.py'
      ],

      include_package_data=True,
      zip_safe=False)



"""Decorators meethods to support broadcast controller running functions on multiple devices"""
import logging
import serial

from .exceptions import HeatmiserResponseError, HeatmiserControllerTimeError

class ListWrapperClass(object):
    """Class to provide mutable list as decorator argument"""
    def __init__(self):
        self._storedlist = None

    @property
    def list(self):
        """I'm the 'x' property."""
        return self._storedlist

    @list.setter
    def list(self, value):
        self._storedlist = value

    @list.deleter
    def list(self):
        self._storedlist = None

def run_function_on_all(liststore):
    """Decorator to allow a class method to be run on all objects in a list"""
    def wraps(func):
        """Decorator internal"""
        def inner(self, *args, **kwargs):
            """Decorator internal"""
            if liststore.list is None:
                raise ValueError("liststore contains no list")
            logging.info("All running %s for %i controllers"%(func.__name__, len(liststore.list)))
            func(self, *args, **kwargs)
            results = [None] * len(liststore.list)
            lasterror = None
            for index, obj in enumerate(liststore.list):
                try:
                    results[index] = getattr(obj, func.__name__)(*args, **kwargs)
                except (HeatmiserResponseError, serial.SerialException, HeatmiserControllerTimeError) as err:
                    logging.warn("C%i %s failed due to %s"%(obj._address, func.__name__, str(lasterror)))
                    lasterror = err
                    continue

            if all(result is None for result in results):
                raise HeatmiserResponseError("All failed, last error was %s"%(str(lasterror)))

            return results

        return inner
    return wraps
